/*
** Module   : ComboBoxGuiImpl.java
** Abstract : GUI combo-box implementation
**
** Copyright (c) 2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 EVL 20150421 Initial version.
** 002 GES 20150508 Eliminated AWT dependencies.
** 003 EVL 20150514 Adding GUI specific keyboard and mouse default exit events.  Turning on
**                  tooltip engine.  Adding consistent cleanup on drop-down exit for GUI mode.
**                  Adding scaling factor to get rid of the hardcoding adjustments.  Moving sets
**                  location of the drop-down button to later stage, just before drawing to
**                  properly support dynamic combo-box.
** 004 EVL 20150719 Fix for issue when initial value displayed in entry field when nothing yet is
**                  selected.  We can use selected flag to detect the difference.  Adding another
**                  scaling option to compute required widget height.  Fix for getting pixel width
**                  and height attribute.  Minor painting details coord issues.  TAB/BACK_TAB
**                  event processing fix.  Need to recalc button location before every painting
**                  because the width can be changed every time with width-* attribute.  Clean up
**                  the scaling compute method.
** 005 EVL 20150814 Adding helper method to provide row height to parts of the combo-box.  If the
**                  width or height was set explicitly - return it immediately.
** 006 EVL 20150907 Reworked drop-down button drawing to use SVG engine to get proper image.
** 007 HC  20150831 Scrolling support overhaul.
** 008 CA  20150910 Added mouse actions.
** 009 IAS 20150924 AUTO-RESIZE and TOOLTIP attributes support
** 010 CA  20150925 Made exitDropDown() public, it needs to be called if the drop-down is
**                  notified that is losing focus.  Fixed multi-window drawing issues caused by 
**                  concurrent drawing access to the GuiDriver: all access to the driver is done
**                  exclusively, each thread taking and releasing ownership for the duration of
**                  the drawing operation or other GUI API invocation which requires access to
**                  the underlying physical window.
** 011 HC  20151013 Changes to extend GUI multi-window focus management and ACTIVE-WINDOW system
**                  handle processing to modal windows.
** 012 EVL 20151014 Fix for button background color issue. Fix for incorrect displaying the value
**                  instead of the label om initial start. 
** 013 HC  20151029 Fixed drop-down scroll list width calculation, it must not be smaller than
**                  the parent's width.
**         20151103 Improved GUI COMBO-BOX layout.
**         20151110 GUI combo-box fixes.
**         20151112 Implemented THREE-D runtime support.
**     EVL 20151118 Adding override for getDataEntryReturn().  It should always return FALSE in
**                  GUI mode ignoring SESSION:DATA-ENTRY-RETURN value.  The combo-box does not
**                  react to the ENTER key in GUI.  Key processsing fixes for GUI mode.
*/

package com.goldencode.p2j.ui.client.gui;

// this usage of AWT events is intentional, they are being used separately from the
// AWT/Swing event model itself; no other AWT/Swing usage is occurring in this class
import java.awt.event.MouseEvent;
import java.util.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.ComboBoxConfig.Mode;
import com.goldencode.p2j.ui.chui.ThinClient;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.format.*;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.gui.GuiFontResolver.FontCache;
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.p2j.ui.client.ScrollBar.Bars;
import com.goldencode.p2j.util.*;
import com.goldencode.util.*;

/**
 * Specific GUI implementation of COMBO-BOX widget.
 */
@SuppressWarnings("unchecked")
public class ComboBoxGuiImpl
extends ComboBox<GuiOutputManager>
implements AutoResizable<ComboBoxConfig>
{
   /** Shift in X to draw text inside entry. */
   public static final int TEXT_SHIFT = 4;
   
   /** Shift in X and Y to draw entry field for DROP-DOWN and SIMPLE modes. */
   public static final int ENTRY_SHIFT = 2;
   
   /** Dimension delta to draw entry field for DROP-DOWN and SIMPLE modes. */
   public static final int ENTRY_DELTA = ENTRY_SHIFT + 1;
   
   /** Screen driver */
   private GuiDriver gd;

   /** Output manager */
   private GuiOutputManager screen;
   
   /** Coordinate converter */
   private CoordinatesConversion cc;

   /** Color resolver */
   private GuiColorResolver gc;
   
   /** Font resolver */
   private GuiFontResolver gf;
      
   /** Selected item box color. */
   private ColorRgb selectionBack;

   /** Selected dotted line color. */
   private ColorRgb selectionDots;

   /** Light shadow box color. */
   private ColorRgb lightShadow;

   /** Dark shadow box color. */
   private ColorRgb darkShadow;
   
   /** Very dark shadow box color. */
   private ColorRgb veryDarkShadow;
   
   /** Drop down button background color. */
   private ColorRgb btnBack;
   
   /** Pressed button state flag. */
   private boolean pressed = false;
   
   /** Image to draw drop down button. */
   private ImageWrapper<?> imgButtEn = null;
   
   /** Image to draw drop down button in disabled state. */
   private ImageWrapper<?> imgButtDis = null;
   
   /** Editable entry field for SIMPLE and DROP-DOWN modes. */
   private EntryFieldGuiImpl entryField = null;
   
   /** The list for SIMPLE combo-box mode. */
   private ScrollableSelectionListGuiImpl simpleList = null;
   
   /** The list for SIMPLE combo-box mode. */
   private ScrollPaneGuiImpl simpleListPane = null;
   
   /** Name of the image to use for drop-down list activation button. */
   private static final String DROP_DOWN_BUTTON_NAME = "sb-arr-en-down";
   
   /** Name of the image to use for drop-down list activation button disabled sate. */
   private static final String DROP_DOWN_BUTTON_NAME_INS = "sb-arr-dis-down";
   
   /** Size of the drop-down button image. */
   private static final int DROP_DOWN_BUTTON_SIZE = 16;

   /**
    * Special constructor used to restore component using given configuration.
    * 
    * @param   id
    *          The widget's ID.
    * @param   cfg
    *          Configuration data.
    */
   public ComboBoxGuiImpl(WidgetId id, WidgetConfig cfg)
   {
      super(id, (ComboBoxConfig) cfg);
   }
   
   /**
    * Replacement API for the logic which should exist in the widget's c'tor, but the logic 
    * can't be executed as it depends on the widget being registered into the official registry.
    * <p>
    * This API is called after the widget is created and added to the registry.
    * 
    * @param    id
    *           The widget's ID.
    * @param    wcfg
    *           The config used to initialize this widget.
    */
   @Override
   public void initialize(WidgetId id, ComboBoxConfig wcfg)
   {
      super.initialize(id, wcfg);
      
      // screen driver
      if (gd == null)
      {
         gd = (GuiDriver) OutputManager.getDriver();
      }
      
      // output manager
      screen = screen();
      cc = screen.coordinates();
      
      // font resolver
      if (gf == null)
      {
         gf = new GuiFontResolver(WindowManager.resolveWindow(this), gd, config(), false);
      }
      
      // initialize the shadow colors
      lightShadow    = gd.getSysColor(SysColor.COLOR_WINDOW);
      darkShadow     = gd.getSysColor(SysColor.COLOR_3DSHADOW);
      veryDarkShadow = gd.getSysColor(SysColor.COLOR_3DDKSHADOW);
      selectionBack  = gd.getSysColor(SysColor.COLOR_HIGHLIGHT);
      btnBack        = gd.getSysColor(SysColor.COLOR_3DFACE);
      
      // the color below is negative to selectionBack
      selectionDots  = new ColorRgb(255 - selectionBack.getRed(),
                                    255 - selectionBack.getGreen(),
                                    255 - selectionBack.getBlue());

      // create required images to draw drop-down button
      if (config().mode == ComboBoxConfig.Mode.DROP_DOWN ||
          config().mode == ComboBoxConfig.Mode.DROP_DOWN_LIST)
      {
         imgButtEn = VectorGraphicsHelper.getScalableImage(DROP_DOWN_BUTTON_NAME, gd,
                                                           DROP_DOWN_BUTTON_SIZE,
                                                           DROP_DOWN_BUTTON_SIZE);
         imgButtDis = VectorGraphicsHelper.getScalableImage(DROP_DOWN_BUTTON_NAME_INS, gd,
                                                            DROP_DOWN_BUTTON_SIZE,
                                                            DROP_DOWN_BUTTON_SIZE);
      }
   }

   /**
    * The method is called after the configuration associated with 
    * the implementor changes. The parameter points to a reference
    * representing the original configuration state. This reference
    * can be used to detect what configuration fields changed and so
    * optimize any processing related to the configuration change.
    * 
    * @param   beforeUpdate
    *          Config reference capturing the config state before 
    *          the owner's config was modified.
    */
   @Override
   public void afterConfigUpdate(ComboBoxConfig beforeUpdate)
   {
      super.afterConfigUpdate(beforeUpdate);
      
      if (!StringHelper.equal(config.format, beforeUpdate.format) ||
          config.font != beforeUpdate.font)
      {
         updateSize(beforeUpdate);
      }
   }
   
   /**
    * Draw the combo-box. Called by this ComboBox's parent container.
    */
   @Override
   public void draw()
   {
      // get location and size
      NativePoint p = physicalLocation();
      final NativeDimension d = physicalDimension();

      // update entry field for SIMPLE and DROP-DOWN modes
      if (entryField != null)
      {
         entryField.setPhysicalLocation(p.x + ENTRY_SHIFT, p.y + ENTRY_SHIFT);
         
         // and set up initial value for entry field
         if (isSelected())
         {
            entryField.setValue(new character(model().selected()));
         }
         else
         {
            BaseDataType initial = getValue();
            if (initial != null && !initial.isUnknown())
            {
               entryField.setValue(new character(getInitialScreenText(initial)));
            }
         }
      }
      
      // layout simple mode list
      if (simpleListPane != null)
      {
         ThinClient.getInstance().independentEventDrawingBracket(this, new Runnable()
         {
            public void run()
            {
               layoutSimpleList();
            }
         });
      }
      
      // combo-box options need to be refreshed before every painting
      final ComboBoxConfig cbbcfg = config();
      gc = new GuiColorResolver(gd, cbbcfg, SysColor.COLOR_MENU, SysColor.COLOR_BTNTEXT);
      
      // determine colors to use
      gc.colors();
      
      // resolve font
      final String text = getLabelText();
      final FontCache fntCache = gf.font();

      // set up clipping rectangle
      NativeRectangle clip = simpleListPane != null
         ? clipRectangle(p, new NativeDimension(d.width + 3, d.height +
                                                simpleListPane.physicalDimension().height + 6))
         : clipRectangle(p, new NativeDimension(d.width - 1, d.height));
      
      // clippable drawing start 
      gd.draw(p, clip, new Runnable()
      {

         @Override
         public void run()
         {
            // inside this call widget origin is always (0,0)
            NativePoint po = new NativePoint(0, 0);

            boolean focused = hasFocus() && isEnabled();

            // text in DROP-DOWN-LIST mode
            if (config().mode == ComboBoxConfig.Mode.DROP_DOWN_LIST)
            {
               drawEntryField(po, d, fntCache.font, focused);
            }
            // in SIMPLE or DROP-DOWN modes the entry field drawing
            // is happening from the frame the entry field belongs to

            // the field border            
            drawFieldBorder(po, d);

            // and drop-down button for modes except simple
            if (config().mode != ComboBoxConfig.Mode.SIMPLE)
            {
               drawButton(po, d, pressed);
            }

            // draw drop-down in SIMPLE mode border
            if (config().mode == ComboBoxConfig.Mode.SIMPLE && simpleListPane != null)
            {
               NativePoint slP = new NativePoint(0, d.height + 2);
               NativeDimension slDim = simpleListPane.physicalDimension();
               slDim.width += 3;
               slDim.height += 3;
               drawFieldBorder(slP, slDim);
            }
            // in SIMPLE mode the list drawing is happening from the frame the list belongs to
         }
         
      });
      
   }

   
   /**
    * Sets position of terminal's cursor to the first char of the combo-box.
    */
   @Override
   public void drawCaret()
   {
      // this check is required since the method can be called from multiple
      // drivers
      
      // nothing to do in GUI
   }

   /**
    * Obtain the input focus.
    */
   @Override
   public void requestFocus()
   {
      if (isEnabled())
      {
         // generate the FOCUS_GAINED event
         if (entryField != null)
         {
            entryField.requestFocus();
         }
         else
         {
            super.requestFocus();
         }
      }
   }

   /**
    * Notification of a mouse pressed event occurred for this widget.  Redraw combo-box in pressed
    * state and request focus.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mousePressed(MouseEvent e)
   {
      // push button effect
      if (isEnabled() && !pressed)
      {
         requestFocus();
         pressed = true;
         // drop-down is opening on mouse press
         if (config.mode != ComboBoxConfig.Mode.SIMPLE)
         {
            if (dropDownActive)
            {
               // second click to drop-down button or area dismisses the active drop-down
               // simulate ESC for scrollable item list to exit drop-down
               KeyInput ki = new KeyInput(((ScrollPaneGuiImpl)dropDown.getList()).getChild(),
                                          EventType.KEY_PRESSED,
                                          Key.VK_ESCAPE);
               EventManager.postEvent(ki);
               // do not handle further pressings
               pressed = false;
            }
            else
            {
               // activate drop-down
               activate();
            }
         }
      }
   }

   /**
    * Notification of a mouse released event occurred for this widget.  Redraw combo-box in
    * released state.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseReleased(MouseEvent e)
   {
      // release button effect
      if (isEnabled() && pressed)
      {
         pressed = false;
         repaint();
      }
   }

   /**
    * Process key events that have been generated by this component.
    * 
    * @param    keyEvent
    *           The key event to be handled.
    */
   @Override
   public void processKeyEvent(KeyInput keyEvent)
   {
      if (!isEnabled())
      {
         return;
      }
      
      // this event is generated on combo-box parts to invoke a 4GL triggers
      // we need to consume and ignore it because triggers should already be called
      if (keyEvent.actionCode() == Keyboard.SE_VALUE_CHANGED)
      {
         keyEvent.consume();
         return;
      }
      
      if (!dropDownActive)
      {
         int key = keyEvent.actionCode();
         
         // consider focus move events first
         if (key == Keyboard.KA_TAB || key == Keyboard.KA_BACK_TAB)
         {
            super.processKeyEvent(keyEvent);
         }
         else if (entryField != null && key != Keyboard.SE_VALUE_CHANGED)
         {
            entryField.processKeyEvent(keyEvent);
         }
         else
         {
            int newIndx = ScrollableList.getFirstCharSelection(keyEvent, model(),
                                                         model().getSelectedIndex());
            if (newIndx >= 0)
            {
               // select the new item
               setNewSelection(newIndx);
               // and consume event
               keyEvent.consume();
            }
            else
            {
               super.processKeyEvent(keyEvent);
            }
         }
      }
      else if (dropDown != null)
      {
         if (keyEvent.actionCode() == Keyboard.SE_LEAVE)
         {
            KeyInput ki = new KeyInput(((ScrollPaneGuiImpl)dropDown.getList()).getChild(),
                                       EventType.KEY_PRESSED,
                                       Key.VK_ESCAPE);
            EventManager.postEvent(ki);
            // do not handle further pressings
            pressed = false;
         }
         else
         {
            dropDown.processKeyEvent(keyEvent);
         }
      }
      else
      {
         super.processKeyEvent(keyEvent);
      }
   }
   
   /**
    * Process the action associated with the given key in an independent drawing bracket.  In
    * GUI combo-box the precessing is highly dependend on the current mode.
    * <p>
    * This must be called only in the context of widget-level event processing.
    * 
    * @param    keyEvent
    *           The key to be processed.
    */
   @Override
   public void processSystemKey(final KeyInput keyEvent)
   {
      // get the key code
      int key = keyEvent.actionCode();
   
      // if the key is not in GUI specific key set - call superclass for any mode
      if (key != Keyboard.KA_CURSOR_DOWN && key != Keyboard.KA_CURSOR_UP &&
          key != Keyboard.KA_CURSOR_LEFT && key != Keyboard.KA_CURSOR_RIGHT &&
          key != Keyboard.KA_PAGE_DOWN && key != Keyboard.KA_PAGE_UP &&
          key != Keyboard.KA_HOME && key != Keyboard.KA_END)
      {
         // check if we press GUI combo-box drop-down opener
         int keyCode = keyEvent.keyCode();
         // alternative GUI drop-down opening by keyboard
         if (keyCode == Key.VK_ALT_CURSOR_DOWN || keyCode == Key.VK_ALT_CURSOR_UP)
         {
            activate();
         }
         else
         {
            super.processSystemKey(keyEvent);
         }
      }
      // below is the GUI specific keys processing depending on mode
      else if (config.mode == ComboBoxConfig.Mode.SIMPLE)
      {
         // the mode is simple - route processing to the list for all keys
         simpleList.processKeyEvent(keyEvent);
      }
      // the mode is DROP-DOWN or DROP-DOWN-LIST
      // the keys below change the current selection not opening drop-down part
      else
      {
         // get the currently selected index
         int currIndex = model().getSelectedIndex();
         int newIndexCandidate = currIndex; 

         // change index depending on key types
         if (key == Keyboard.KA_CURSOR_DOWN || key == Keyboard.KA_CURSOR_RIGHT)
         {
            newIndexCandidate++;
         }
         else if (key == Keyboard.KA_CURSOR_UP || key == Keyboard.KA_CURSOR_LEFT)
         {
            newIndexCandidate--;
         }
         else if (key == Keyboard.KA_PAGE_DOWN)
         {
            newIndexCandidate += getInnerLines();
         }
         else if (key == Keyboard.KA_PAGE_UP)
         {
            newIndexCandidate -= getInnerLines();
         }
         else if (key == Keyboard.KA_HOME)
         {
            newIndexCandidate = 0;
         }
         else if (key == Keyboard.KA_END)
         {
            newIndexCandidate = model().size() - 1;
         }
         
         // check the edge conditions
         if (newIndexCandidate < 0)
         {
            newIndexCandidate = 0;
         }
         else if (newIndexCandidate >= model().size())
         {
            newIndexCandidate = model().size() - 1;
         }

         // nothing is changed, return
         if (newIndexCandidate == currIndex)
         {
            return;
         }
         
         // start actual intenal model change
         final int newIndex = newIndexCandidate;
         
         // select the next item
         ThinClient.getInstance().independentEventDrawingBracket(this, new Runnable()
         {
            @Override
            public void run()
            {
               setNewSelection(newIndex);
            }
         });
      }
   }

   /**
    * Process the action associated with the given focus movement key in an independent drawing
    * bracket.  Currently it is <code>TAB</code> or <code>BACK_TAB</code>.
    * <p>
    * This must be called only in SIMPLE or DROP-DOWN mode either from entry field or from
    * attached selection list.
    * 
    * @param    keyEvent
    *           The key to be processed, <code>TAB</code> or <code>BACK_TAB</code>.
    */
   public void processFocusMoveKey(KeyInput keyEvent)
   {
      // set the current focus to the main combo-box part
      parent().setFocus(this);
      // call superclass to do the rest
      super.processSystemKey(keyEvent);
   }
   
   /**
    * Get widget screen value.
    * 
    * @return  Current widget screen value.
    */
   @Override
   public BaseDataType getScreenValue()
   {
      return entryField != null ? new character(entryField.getScreenValue(true))
                                : super.getScreenValue();
   }

   /**
    * Shows or hides the combo-box and its label.
    *
    * @param    visible_
    *           <code>true</code> to make the widget visible, 
    *           <code>false</code> to hide the widget.
    */
   @Override
   public void setVisible(boolean visible_)
   {
      super.setVisible(visible_);
      
      if (visible_)
      {
         // create fill-in for SIMPLE and DROP_DOWN modes
         if (entryField == null &&
             (config().mode == ComboBoxConfig.Mode.SIMPLE || 
              config().mode == ComboBoxConfig.Mode.DROP_DOWN))
         {
            createEntryField();
         }

         // for simple mode we need to create permanent drop-down
         if (config().mode == ComboBoxConfig.Mode.SIMPLE && simpleListPane == null)
         {
            createSimpleList();
         }
      }
      
      if (entryField != null)
      {
         entryField.setVisible(visible_);
      }
      
      if (simpleListPane != null)
      {
         simpleListPane.setVisible(visible_);
      }
   }
   
   /**
    * @see com.goldencode.p2j.ui.client.widget.Widget#height()
    */
   @Override
   public double height()
   {
      if (config.clientHeightChars == BaseConfig.INV_COORD)
      {
         updateSize(null);
      }
      
      return config.clientHeightChars != BaseConfig.INV_COORD ?
             config.clientHeightChars : config.heightChars;
   }
   
   /**
    * @see com.goldencode.p2j.ui.client.widget.Widget#width()
    */
   @Override
   public double width()
   {
      if (config.clientWidthChars == BaseConfig.INV_COORD)
      {
         updateSize(null);
      }
      
      return config.clientWidthChars != BaseConfig.INV_COORD ?
             config.clientWidthChars : config.widthChars;
   }
   
   /**
    * This method is invoked when selection is changed.
    */
   @Override
   public void changed()
   {
      // call common code first
      super.changed();

      // then GUI specific
      if (config().mode == ComboBoxConfig.Mode.SIMPLE && simpleList != null)
      {
         setNewSelection(simpleList.index());
      }
      else if (dropDownActive && dropDown != null)
      {
         setNewSelection(dropDown.getSelectedIndex());
      }
   }

   /**
    * Enable/disable widget.
    * 
    * @param   enabled
    *          <code>TRUE</code> to make the widget enabled, 
    *          <code>FALSE</code> to disable the widget.
    */
   @Override
   public void setEnabled(boolean enabled)
   {
      // set flag for main part
      super.setEnabled(enabled);

      // set flag for possible entry filed
      if (entryField != null)
      {
         entryField.setEnabled(enabled);
      }

      // set flag for possible simple list
      if (simpleList != null)
      {
         simpleList.setEnabled(enabled);
      }
      
      if (simpleListPane != null)
      {
         simpleListPane.setEnabled(enabled);
      }

      // and post value changed event
      repaint();
   }

   /**
    * Check if widget has focus.
    * 
    * @return  <code>true</code> if widget has focus.
    */
   @Override
   public boolean hasFocus()
   {
      // check the combo-box itself for focus
      boolean cbbFocused = super.hasFocus();
      
      return config().mode == ComboBoxConfig.Mode.SIMPLE &&
                simpleList != null && entryField != null
                   ? cbbFocused || simpleList.hasFocus() || entryField.hasFocus()
                   : config().mode == ComboBoxConfig.Mode.DROP_DOWN && entryField != null 
                      ? cbbFocused || entryField.hasFocus() || dropDownActive : cbbFocused;
   }
   
   /**
    * Gets the current row height depending on choosen font.
    * 
    * @return  The row height value in pixels.
    */
   public int getRowHeight()
   {
      int font = gf.resolveFontNum();
      return FontManager.getFontHeight(WindowManager.resolveWindow(this), font, gd) + 2;
   }
   
   /**
    * Destroy this widget.  Cleans up all including widgets before destruction.
    */
   public void destroy()
   {
      // deregister possible SIMPLE list components
      if (simpleListPane != null)
      {
         OutputManager.instance().getRegistry().removeWidgetsRecursive((Widget) simpleListPane);
      }

      // call superclass for main destruction
      super.destroy();
   }

   /**
    * Posts value changed event to the queue.  This should be done in GUI mode after any changes
    * in entry field for SIMLPLE and DROP_DOWN combo-box modes.
    */
   public void entryFieldChanged(String newValue)
   {
      // put an ActionEvent onto the system event queue
      EventManager.postEvent(new ActionEvent(this, newValue));
      
      // honor any Progress triggers
      KeyInput vchng = new KeyInput(this, EventType.KEY_PRESSED, Keyboard.SE_VALUE_CHANGED, true);
      EventManager.postEvent(vchng);
      
      ConfigHelper.setModified(config, true);
   }

   /**
    * Exit the drop-down (the "selection" mode of the combo-box).  GUI specific cleanup.
    */
   @Override
   public void exitDropDown()
   {
      // call generic cleanup from the superclass
      super.exitDropDown();

      // clean up registry from drop-down widgets added in GUI code
      OutputManager.instance().getRegistry().removeWidgetsRecursive((Widget) dropDown);
   }
   

   /**
    * Update widget's size after config update
    * 
    * @param beforeUpdate
    *        previous value of the widget's config
    */
   @Override
   public void updateSize(ComboBoxConfig beforeUpdate)
   {
      if (gd == null)
      {
         return;
      }
      
      // current combo-box window
      Window<?> window = parent(Window.class);
      if (window == null)
      {
         // the widget is not attached, just return
         return;
      }
      
      CoordinatesConversion cc = screen().coordinates();
      
      // get font resolver
      if (gf == null)
      {
         gf = new GuiFontResolver(window, gd, config(), false);
      }
      
      // get font to use in combo-box
      int font = gf.resolveFontNum();
      
      double inputHeight = config.mode == Mode.SIMPLE ? config.heightChars : 1;
      config.clientHeightPixels = cc.heightToNative(1, window.getId(), font) + 8;
      
      if (config.widthChars != 0)
      {
         // the explicitly assigned width is the exact width of the widget,
         // ignore format in this case.
         // the width in this case is converted to physical size using
         // the default font.
         config.clientWidthChars = config.widthChars;
         return;
      }
      
      double charWidth = getFormatLength();
      
      // TODO: variable button size
      // the extra 28 pixels cover the button, borders and padding,
      // the value was determined experimentally
      config.clientWidthPixels = cc.widthToNative(charWidth, window.getId(), font) + 28;
   }

   /**
    * Get the AUTO-RESIZE status
    * 
    * @return <code>true</code> is AUTO-RESIZE is enabled
    */
   @Override
   public boolean isAutoResize()
   {
      return config.autoResize;
   }

   /**
    * Returns the list for SIMPLE combo-box mode.
    *
    * @return   The list instance.
    */
   public Object getList()
   {
      return simpleList;
   }

   /**
    * Determine the mouse actions processed by this widget.
    * 
    * @return   See above.
    */
   @Override
   protected int[] mouseActions()
   {
      return new int[] { MouseEvent.MOUSE_PRESSED, MouseEvent.MOUSE_RELEASED };
   }
   
   /**
    * Posts value changed event to the queue.  This should be done in ChUI and GUI mode after
    * closing drop-down or in GUI mode after pressing UP/DOWN keys with focused combo-box.
    */
   @Override
   protected void valueChangedCommon()
   {
      // set up entry field if exists
      if (entryField != null)
      {
         String newText = model().selected();
         // set new value
         entryField.setValue(new character(newText));
         // set up selection range to the full text selection
         entryField.config().selectionStart = 0;
         entryField.config().selectionEnd = newText.length();
      }
      
      // and call superclass
      super.valueChangedCommon();
      
      // clear changed flag to avoid duplications
      valueChanged = false;
      // change the selected flag
      selected = true;
   }
   
   /**
    * Initializes the event set to return.  GUI specifics,
    * 
    * @return  The keyboard event to be used to exit drop-down.
    */
   @Override
   protected Set<String> initEventsSet()
   {
      // call common version
      Set<String> setGuiResult = super.initEventsSet();

      // add GUI specific messages
      setGuiResult.add("TAB");
      setGuiResult.add("BACK-TAB");
      setGuiResult.add("LEAVE");
      
      return setGuiResult;
   }
   
   /**
    * Adding mouse clicks as exit events.  GUI specifics.
    * 
    * @param   listToChange
    *          The event list to modify.
    */
   @Override
   protected void addUISpecificEvents(EventList listToChange)
   {
      // call common version first
      super.addUISpecificEvents(listToChange);

      // mouse click on anywhere dismisses drop-down
      listToChange.addEvent("MOUSE-MENU-DOWN", true);
      listToChange.addEvent("MOUSE-EXTEND-DOWN", true);
      
      // chaining current event list from ThinClient enclosing event loop
      EventList elCurr = ThinClient.getInstance().getCurrentEventList();
      // chain only events for the given combo-box instance
      if (elCurr != null)
      {
         listToChange.setChain(elCurr.getNonExitEventsForWidget(getId().asInt()));
      }

      // enable drop-down painting
      ((DropDownGuiImpl)dropDown).enableDraw();
   }

   /**
    * Convenience method to get access to SESSION:DATA-ENTRY-RETURN attribute.
    * 
    * @return  Always <code>FALSE</code> for GUI mode.
    */
   @Override
   protected boolean getDataEntryReturn()
   {
      return false;
   }
   
   /**
    * Clean up the main registry pool from drop-down widgets when it is closing, if the widget is
    * a container, make recursive call.
    */
   private void createEntryField()
   {
      // get ID for new widget
      WidgetId efId = WidgetId.nextID();

      // create config for entry field, we use fill-in config for this widget,
      // meaning no server-side counterpart for this widget
      FillInConfig finCfg = new FillInConfig(efId);

      // inherit some options from combo-box config
      finCfg.labels = false;
      finCfg.frameId = config.frameId;
      finCfg.dataType = config.dataType;
      finCfg.setDynamicFormat(config.format);

      // create entry field
      entryField = new EntryFieldGuiImpl(finCfg, this);
      // this is required to properly initialize entry field
      GuiOutputManager.instance().getRegistry().addWidget(efId, entryField);
      // initialize entry field
      entryField.initialize(efId, finCfg);
      // add widget config to config manager
      ConfigManager.getInstance().addWidgetConfig(finCfg);

      // add the widget to the frame the combo-box belongs to
      Frame<GuiOutputManager> frCombo = (Frame<GuiOutputManager>) UiUtils.locateFrame(this);
      frCombo.addWidget(entryField);
   }
   
   /**
    * Creates the list for SIMPLE mode of the combo-box.  This part will permanently exist for
    * the combo-box life time.
    */
   private void createSimpleList()
   {
      // create the drop-down box
      simpleList = screen().getFactory().createScrollableList(this, this.model());

      // register it as a new frame
      simpleListPane = screen().getFactory().createScrollPane(simpleList);

      // set height, width and position of the simlpe itself
      // account for the border
      simpleList.setColumns(width() - simpleListPane.getRootPanel().insets().widthDelta());

      // set up visible rows
      int rows = config.innerLines;
      
      if (rows <= 0)
      {
         int numItems = model().size();
         
         // ensure that we always have at least 1 row
         rows = (numItems > 0) ? numItems : 1;
      }
      simpleList.setVisibleRows(rows);
      // force the initial selection
      int idx = model().calcSelection(config.items);
      
      simpleList.select(idx);
      
      // set up initial top row
      simpleList.setTopRow(idx);
     
      // set up notifications
      simpleList.addSelectionListener(this);

      // init scrollbars, visible on demand
      simpleListPane.setScrollBars(Bars.VERTICAL);
         
      // adding to the combo-box parent frame
      Frame<GuiOutputManager> frCombo = (Frame<GuiOutputManager>) UiUtils.locateFrame(this);
      frCombo.addWidget(simpleListPane);

      // adding to the registry, no mouse event will be available if not to register
      OutputManager.instance().getRegistry().addWidgetsRecursive((Widget) simpleListPane);
   }

   /**
    * Layout simple list.
    */
   private void layoutSimpleList()
   {
      int rows = config.innerLines;
      
      double w = width();
      double h = screen().coordinates().heightFromNative(rows * getRowHeight()); 
      simpleListPane.setSize(w, h);

      // set location in character units      
      Point origin = screenLocation();
      origin.y += 1;
      simpleListPane.setLocation(origin);

      // set up physical location just below the combo-box
      NativePoint slLoc = physicalLocation();
      // the shift to draw the list properly
      slLoc.x += 2;
      slLoc.y += physicalDimension().height + 4;
      simpleListPane.setPhysicalLocation(slLoc.x, slLoc.y);
   }
   
   /**
    * Changes the selected value without opening drop-down.
    * 
    * @param   indexNew
    *          The index of the new item to select.
    */
   private void setNewSelection(int indexNew)
   {
      // new value can not be out of the current range
      if (indexNew < 0 || indexNew >= model().size())
      {
         return;
      }
      
      // set up the model with new selection
      model().setSelectedIndex(indexNew);
      model().setSelectedId(indexNew + 1);
      model().select(model().get(indexNew));

      // and post value changed event
      valueChangedCommon();
   }

   /**
    * Draw 3D border around different parts of the combo-box widget.
    * 
    * @param   orig
    *          Combo-box origin.
    * @param   dim
    *          Combo-box size.
    */
   private void drawFieldBorder(NativePoint orig, NativeDimension dim)
   {
      int x1 = orig.x;
      int x2 = orig.x + dim.width - 1;
      int y1 = orig.y;
      int y2 = orig.y + dim.height;

      // drawing the frame around entry field
      gd.setColor(darkShadow);
      gd.drawLine(x1, y1, x1, y2);
      gd.drawLine(x1 + 1, y1, x2, y1);

      gd.setColor(veryDarkShadow);
      gd.drawLine(x1 + 1, y1 + 1, x1 + 1, y2 - 1);
      gd.drawLine(x1 + 2, y1 + 1, x2 - 1, y1 + 1);
      
      gd.setColor(gd.getSysColor(SysColor.COLOR_3DFACE));
      gd.drawLine(x1 + 1, y2 - 1, x2, y2 - 1);
      gd.drawLine(x2, y2 - 1, x2, y1 + 1);
      
      gd.setColor(gd.getSysColor(SysColor.COLOR_WINDOW));
      gd.drawLine(x1, y2, x2 + 1, y2);
      gd.drawLine(x2 + 1, y2, x2 + 1, y1);
   }
   
   /**
    * Draw entry field of the combo-box widget.
    * 
    * @param   orig
    *          Combo-box origin.
    * @param   dim
    *          Combo-box size.
    * @param   fnt
    *          Font to use for combo-box entry field.
    * @param   focused
    *          Is the combo-box focused.
    */
   private void drawEntryField(NativePoint orig,
                               NativeDimension dim,
                               FontDetails<?> fnt,
                               boolean focused)
   {
      // coordinates of the field area
      int w = dim.width - ButtonGuiImpl.PREBUILT_SIZE - 5;
      int h = dim.height - 3; 

      // draw internal background
      if (isThreeD() && !isEnabled())
      {
         gd.setColor(gd.getSysColor(SysColor.COLOR_3DFACE));
      }
      else
      {
         gd.setColor(lightShadow);
      }
      
      gd.drawRect(orig.x + 2, orig.y + 2, w, h );
      if (focused)
      {
         gd.setColor(selectionBack);
      }
      gd.fillRect(orig.x + 3, orig.y + 3, w - 1, h - 1);
      
      // get the text to draw with field
      String display = null;
      
      if (isSelected())
      {
         display = model().selected();
      }
      else
      {
         BaseDataType initial = getValue();
         if (initial != null && !initial.isUnknown())
         {
            display = getInitialScreenText(initial);
         }
      }
      
      if (display == null)
         display = new String("");

      // limit the data portion of the field to the maximum allowed size
      int limit = getFormatLength();
      
      if (display.length() > limit)
      {
         display = display.substring(0, limit);
      }

      // draw entry field text
      gd.setGuiFont(fnt);
      // left-center alignment
      int xt = orig.x + TEXT_SHIFT;
      int yt = orig.y + 2 + (dim.height + fnt.pointSize)/2;
      gd.setFontStyle(fnt.style);
      // enabled state controls text drawing
      if (isEnabled())
      {
         if (focused)
         {
            gd.setColor(lightShadow);
         }
         else
         {
            gd.setColor(gc.fgColor);
         }
         gd.drawString(display, xt, yt);
      }
      else
      {
         gd.setColor(lightShadow);
         gd.drawString(display, xt + 1, yt + 1);
         gd.setColor(darkShadow);
         gd.drawString(display, xt, yt);
      }

      // draw dotted rectangle around focused button text
      if (focused)
      {
         // black background for dotted line
         gd.setColor(gc.fgColor);
         gd.drawRect(orig.x + 3, orig.y + 3, w - 2, h - 2);
         // set dotted line style
         gd.setLineStroke(LineStroke.DOTTED_SMALL);
         gd.setColor(selectionDots);
         gd.drawRect(orig.x + 3, orig.y + 3, w - 2, h - 2);
         // restore default line style
         gd.setLineStroke(LineStroke.DEFAULT);
      }
   }
   
   /**
    * Gets the initially setted screen value.
    * 
    * @param   initVal
    *          The number of the list items pair or entry for the list item.
    * 
    * @return  The string representation of the initial value.
    */
   private String getInitialScreenText(BaseDataType initVal)
   {
      // first try to scan by value index
      String ndx = initVal.toStringMessage();
      String res = null;

      // scan the config for items
      for (ControlSetItem itm : config.items)
      {
         if (itm.getValue().toStringMessage().equals(ndx))
         {
            // found the match exiting
            res = itm.getLabel().toStringMessage();
            break;
         }
      }

      // initial value is not index, try to get the value as is
      if (res == null)
      {
         res = initVal.toStringMessage();
      }
      
      return res;
   }

   /**
    * Draw drop down button of the combo-box widget.
    * 
    * @param   orig
    *          Combo-box origin.
    * @param   dim
    *          Combo-box size.
    * @param   pressed
    *          Flag indicating the button state, pressed or released.
    */
   private void drawButton(NativePoint orig, NativeDimension dim, boolean pressed)
   {
      // simple check to prevent NPE
      if (imgButtEn == null || imgButtDis == null)
      {
         return;
      }
      
      // coordinates of the button area
      int x1 = orig.x + dim.width - ButtonGuiImpl.PREBUILT_SIZE - 2;
      int x2 = orig.x + dim.width;
      int y1 = orig.y + 2;
      int y2 = orig.y + dim.height;

      // draw the light frame around image
      gd.setColor(btnBack);
      gd.fillRect(x1, y1, x2 - x1, y2 - y1);
      
      // center the image
      int imageY = y1 + (y2 - y1 - DROP_DOWN_BUTTON_SIZE) / 2;
      
      // disabled state has separate inage to draw  
      if (pressed)
      {
         // for pressed state need 1 pixel right-down shift 
         gd.drawImage(imgButtEn, x1 + 1, imageY + 1,
                      DROP_DOWN_BUTTON_SIZE, DROP_DOWN_BUTTON_SIZE, 0, 0,
                      false, false, false, false);
         
         // draw small button frame
         gd.setColor(darkShadow);
         gd.drawRect(x1, y1, x2 - x1 - 1, y2 - y1 - 1);
      }
      else
      {
         // determine what image to draw
         ImageWrapper<?> imgReleased = null;
         if (isEnabled())
         {
            imgReleased = imgButtEn;
         }
         else
         {
            imgReleased = imgButtDis; 
         }
         gd.drawImage(imgReleased, x1, imageY,
                      DROP_DOWN_BUTTON_SIZE, DROP_DOWN_BUTTON_SIZE, 0, 0,
                      false, false, false, false);
         // draw small button frame
         gd.setColor(lightShadow);
         gd.drawLine(x1 + 1, y1 + 2, x1 + 1, y2 - 3);
         gd.drawLine(x1 + 1, y1 + 1, x1 + ButtonGuiImpl.PREBUILT_SIZE - 2, y1 + 1);

         gd.setColor(darkShadow);
         gd.drawLine(x1 + 1, y2 - 2, x1 + ButtonGuiImpl.PREBUILT_SIZE - 1, y2 - 2);
         gd.drawLine(x1 + ButtonGuiImpl.PREBUILT_SIZE - 1, y1 + 1,
                     x1 + ButtonGuiImpl.PREBUILT_SIZE - 1, y2 - 3);

         gd.setColor(veryDarkShadow);
         gd.drawLine(x1, y2 - 1, x1 + ButtonGuiImpl.PREBUILT_SIZE, y2 - 1);
         gd.drawLine(x1 + ButtonGuiImpl.PREBUILT_SIZE, y1,
                     x1 + ButtonGuiImpl.PREBUILT_SIZE, y2 - 2);
      }
   }
   
   /**
    * Returns the format length of the data type the combo-box represents.
    * 
    * @return   Format length.
    */
   private int getFormatLength()
   {
      String       type = config.dataType;
      BaseDataType bdt  = DisplayFormat.instanceOfType(type);
      
      return bdt.formatLength(config.format); 
   }
   
   /**
    * Returns the inner lines value for combo-box.  Used in closed state.
    * 
    * @return   Inner lines value.
    */
   private int getInnerLines()
   {
      // start with config value
      int resIL = config.innerLines;
      
      // second try is to ask the list model
      if (resIL <= 0)
      {
         int mSize = model().size(); 
         resIL = mSize > 0 ? mSize : DropDownGuiImpl.INNER_LINES_MIN;
      }
      
      return resIL;
   }
}
